@@ -19,11 +19,11 @@ along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
  *
- $Source: /usr/src/distr/ftape-1.13/RCS/calibr.h,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/calibr.h,v $
  $Author: bas $
  *
- $Revision: 1.14 $
- $Date: 1994/06/12 16:07:53 $
+ $Revision: 1.18 $
+ $Date: 1995/01/08 14:16:21 $
  $State: Beta $
  *
  *      This file contains a gp calibration routine for
@@ -35,6 +35,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 extern int calibrate( char* name, void(*fun)(int), int* calibr_count, int* calibr_time);
 extern void do_gettimeofday( struct timeval *tv);
 extern int timestamp( void);
+extern int timediff( int t0, int t1);
 
 #endif
 
