@@ -19,11 +19,11 @@ along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
  *
- $Source: /usr/src/distr/ftape-1.13a/RCS/fc-10.h,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/fc-10.h,v $
  $Author: bas $
  *
- $Revision: 1.1 $
- $Date: 1994/06/26 14:10:38 $
+ $Revision: 1.3 $
+ $Date: 1995/01/08 14:16:21 $
  $State: Beta $
  *
  *      This file contains definitions for the FC-10 code
