@@ -15,18 +15,18 @@ You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
- $Source: /usr/src/distr/ftape-1.13b/RCS/fdc-io.c,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/fdc-io.c,v $
  $Author: bas $
  *
- $Revision: 1.31 $
- $Date: 1994/06/30 18:32:14 $
+ $Revision: 1.40 $
+ $Date: 1995/01/08 14:16:21 $
  $State: Beta $
  *
  *      This file contains the low-level floppy disk interface code
  *      for the QIC-40/80 tape streamer device driver.
  */
 
-static char RCSid[] = "$Id: fdc-io.c,v 1.31 1994/06/30 18:32:14 bas Beta $";
+static char RCSid[] = "$Id: fdc-io.c,v 1.40 1995/01/08 14:16:21 bas Beta $";
 
 #include <linux/errno.h>
 #include <linux/sched.h>
@@ -42,6 +42,7 @@ static char RCSid[] = "$Id: fdc-io.c,v 1.31 1994/06/30 18:32:14 bas Beta $";
 #include "ftape-rw.h"
 #include "calibr.h"
 #include "fc-10.h"
+#include "kernel-version.h"
 
 
 /*      Global vars.
@@ -66,9 +67,10 @@ static int fdc_seek_rate = 14;  /* default rate = 2 msec @ 500 Kbps */
 #if defined FDC_BASE || !defined FLOPPY_HACK
 static void (*do_ftape)(void);
 #endif
-static int fdc_fifo_state = 0;  /* original fifo setting */
-static int fdc_lock_state = 0;  /* original lock setting */
-static int fdc_fifo_locked = 0; /* fifo & lock set ? */
+static int fdc_fifo_state;      /* original fifo setting - fifo enabled */
+static int fdc_fifo_thr;        /* original fifo setting - thresshold */
+static int fdc_lock_state;      /* original lock setting - locked */
+static int fdc_fifo_locked = 0; /* has fifo && lock set ? */
 
 
 void
@@ -294,7 +296,7 @@ fdc_issue_command( byte *out_data, int out_count,
   result = fdc_ready_out_wait( 500 /* usec */);
   t1 = timestamp();
   if (result < 0) {
-    TRACEi( 1, "fdc_ready_out_wait failed after:", t0 - t1);
+    TRACEi( 1, "fdc_ready_out_wait failed after:", timediff( t0, t1));
     TRACE_EXIT;
     return result;
   }
@@ -320,6 +322,7 @@ fdc_interrupt_wait( int time) {
   int result = -ETIME;
   int need_cleanup = 0;
   int current_blocked = current->blocked;
+  static int resetting = 0;
   
   if (wait_intr) {
     TRACE( 1, "error: nested call");
@@ -369,8 +372,10 @@ fdc_interrupt_wait( int time) {
   }
   current->blocked = current_blocked; /* restore old state */
   interrupt_seen = 0;		/* clear for next call */
-  if (need_cleanup) {
+  if (need_cleanup & !resetting) {
+    resetting = 1;              /* break infinite recursion if reset fails */
     fdc_reset();
+    resetting = 0;
   }
   TRACE_EXIT;
   return result;
@@ -392,7 +397,11 @@ fdc_motor( int motor)
   } else {
     TRACEx1( 4, "turning motor %d off", unit);
   }
+#ifdef MACH2
+  outb_p( data, fdc.dor2);
+#else
   outb_p( data, fdc.dor);
+#endif
   ftape_sleep( 10 * MILLISECOND);
   TRACE_EXIT;
 }
@@ -403,16 +412,11 @@ fdc_set_data_rate( int rate)
   /* Select clock for fdc, must correspond with tape drive setting !
    * This also influences the fdc timing so we must adjust some values.
    */
-  switch (fdc.type) {
-  case fc10:
-  case i82077:
+  if (fdc.type >= i82077) {
     outb_p( rate & 0x03, fdc.dsr); /* set rate only, keep precomp, reset
                                       and powerdown bits at 0 */
-    break;
-  case i8272:
-  default:
+  } else {
     outb_p( rate, fdc.ccr);     /* set rate */
-    break;
   }
   fdc_data_rate = rate;
   fdc_set_seek_rate( fdc_seek_rate); /* re-adjust for changed clock */
@@ -434,10 +438,20 @@ fdc_reset( void)
   if (ftape_motor) {
     fdc_ctl |= FDC_MOTOR_0 << unit;
   }
+#ifdef MACH2
+  outb_p( fdc_ctl & 0x0f, fdc.dor);
+  outb_p( fdc_ctl, fdc.dor2);
+#else
   outb_p( fdc_ctl, fdc.dor); /* assert reset, keep unit selected */
+#endif
   fdc_usec_wait( 10 /* usec */); /* delay >= 14 fdc clocks */
   fdc_ctl |= FDC_RESET_NOT;
+#ifdef MACH2
+  outb_p( fdc_ctl & 0x0f, fdc.dor);
+  outb_p( fdc_ctl, fdc.dor2);
+#else
   outb_p( fdc_ctl, fdc.dor); /* release reset */
+#endif
   result = fdc_interrupt_wait( 1 * SECOND);
   if (result < 0) {
     TRACE( 1, "missing interrupt after reset");
@@ -460,7 +474,46 @@ fdc_reset( void)
 void
 fdc_disable( void)
 {
+  TRACE_FUN( 8, "fdc_disable");
+  int result;
+  byte cmd1[] = { FDC_CONFIGURE, 0x00, 0x00, 0x00 };
+  byte cmd2[] = { FDC_LOCK };
+  byte cmd3[] = { FDC_UNLOCK };
+  byte stat[ 1];
+  
+  if (fdc_fifo_locked) {
+    result = fdc_issue_command( cmd3, 1, stat, 1);
+    if (result < 0 || stat[ 0] != 0x00) {
+      TRACE( -1, "couldn't unlock fifo, configuration remains changed");
+    } else {
+      cmd1[ 2] = ((fdc_fifo_state) ? 0 : 0x20) + (fdc_fifo_thr - 1);
+      result = fdc_command( cmd1, NR_ITEMS( cmd1));
+      if (result < 0) {
+        TRACE( -1, "couldn't reconfigure fifo to old state");
+      } else if (fdc_lock_state) {
+        result = fdc_issue_command( cmd2, 1, stat, 1);
+        if (result < 0) {
+          TRACE( -1, "couldn't lock old state again");
+        }
+      }
+      TRACEx3( 3, "fifo restored to: %sabled, thresshold %d, %slocked",
+              fdc_fifo_state ? "en" : "dis",
+              fdc_fifo_thr, (fdc_lock_state) ? "" : "not ");
+    }
+    fdc_fifo_locked = 0;
+  }
+#ifdef MACH2
+  outb_p( FTAPE_UNIT & 0x0f, fdc.dor);
+  outb_p( FTAPE_UNIT, fdc.dor2);
+  udelay( 10);
+  outb_p( FDC_RESET_NOT & 0x0f, fdc.dor);
+  outb_p( FDC_RESET_NOT, fdc.dor2);
+#else
   outb_p( FTAPE_UNIT, fdc.dor);
+  udelay( 10);
+  outb_p( FDC_RESET_NOT, fdc.dor);
+#endif
+  TRACE_EXIT;
 }
 
 /*      Specify FDC seek-rate
@@ -469,25 +522,13 @@ int
 fdc_set_seek_rate( int seek_rate)
 {
   byte in[ 3];
-  const int hut = 1;            /* minimal head unload time */
-  const int hlt = 1;            /* minimal head load time */
+  const int hut = 1;            /* minimize head unload time */
+  const int hlt = 1;            /* minimize head load time */
+  const int rates[] = { 250, 300, 500, 1000 };
 
   in[ 0] = FDC_SPECIFY;
-  switch (fdc_data_rate & 0x03) {
-  case 0:
-    in[ 1] = ((16 - (250 * seek_rate) / 500) << 4) | hut;
-    break;
-  case 1:
-    in[ 1] = ((16 - (300 * seek_rate) / 500) << 4) | hut;
-    break;
-  case 2:
-    in[ 1] = ((16 - (500 * seek_rate) / 500) << 4) | hut;
-    break;
-  case 3:
-    in[ 1] = ((16 - (1000 * seek_rate) / 500) << 4) | hut;
-    break;
-  default:
-  }
+  in[ 1] = (((16 - (rates[ fdc_data_rate & 0x03] * seek_rate) / 500) << 4) |
+            hut);
   in[ 2] = (hlt << 1) | 0 /* non_dma */;
   fdc_seek_rate = seek_rate;
 
@@ -553,7 +594,6 @@ fdc_seek( int track)
   int st0, pcn;
 
   ftape_sleep( MILLISECOND);    /* NEEDED to prevent problems */
-
   out[ 0] = FDC_SEEK;
   out[ 1] = FTAPE_UNIT;
   out[ 2] = track;
@@ -690,6 +730,9 @@ setup_fdc_and_dma( buffer_struct* buff, unsigned char operation)
     TRACEx2( 5, "xfer %d sectors to 0x%p", buff->sector_count, buff->ptr);
   } else if (operation == FDC_WRITE || operation == FDC_WRITE_DELETED) {
     dma_mode = DMA_MODE_WRITE;
+    /*  Guess this is the place to turn perpendicular mode on
+     *  when writing QIC-3020 tapes. TO BE IMPLEMENTED
+     */
     TRACEx2( 5, "xfer %d sectors from 0x%p", buff->sector_count, buff->ptr);
   } else {
     TRACE( -1, "bug: illegal operation parameter");
@@ -747,9 +790,10 @@ fdc_fifo_enable( void)
 {
   TRACE_FUN( 8, "fdc_fifo_enable");
   int result = 0;
-  byte cmd0[] = { 0x0e };       /* dumpreg */
-  byte cmd1[] = { 0x13, 0x00, 0x2b, 0x00 }; /* enable fifo, thr = 12 */
-  byte cmd2[] = { 0x94 };       /* lock */
+  byte cmd0[] = { FDC_DUMPREGS };
+  byte cmd1[] = { FDC_CONFIGURE, 0, 0x07, 0 }; /* enable fifo, thr = 8 */
+  byte cmd2[] = { FDC_LOCK };
+  byte cmd3[] = { FDC_UNLOCK };
   byte stat;
   byte reg[ 10];
   int i;
@@ -766,15 +810,26 @@ fdc_fifo_enable( void)
        */
       for (i = 0 ; i < NR_ITEMS( reg) ; ++i) {
         fdc_read( &reg[ i]);
-        TRACEx2( 5, "Register %d = 0x%02x", i, reg[ i]);
+        TRACEx2( 6, "Register %d = 0x%02x", i, reg[ i]);
       }
-      fdc_fifo_state = reg[ 8] & 0x20;
+      fdc_fifo_state = (reg[ 8] & 0x20) == 0;
       fdc_lock_state = reg[ 7] & 0x80;
-      if (fdc_fifo_state) {
-        TRACEx2( 1, "fifo already enabled, thresshold %d%s", reg[ 8] & 0x0f,
-                (fdc_lock_state) ? ", locked" : "");
+      fdc_fifo_thr = 1 + (reg[ 8] & 0x0f);
+      TRACEx3( 3, "original fifo state: %sabled, thresshold %d, %slocked",
+              (fdc_fifo_state) ? "en" : "dis",
+              fdc_fifo_thr, (fdc_lock_state) ? "" : "not ");
+      /*  If fdc is already locked, unlock it first !
+       */
+      if (fdc_lock_state) {
+        fdc_ready_wait( 100);
+        result = fdc_command( cmd3, NR_ITEMS( cmd3));
+        if (result < 0) {
+          TRACE( -1, "FDC unlock command failed, configuration unchanged");
+          result = -EIO;
+        }
       }
-      /*  Enable fifo and set thresshold at 12 bytes for large latency
+      /*  Enable fifo and set thresshold at xx bytes to allow a
+       *  reasonably large latency and reduce number of dma bursts.
        */
       fdc_ready_wait( 100);
       result = fdc_command( cmd1, NR_ITEMS( cmd1));
@@ -790,39 +845,93 @@ fdc_fifo_enable( void)
           result = -EIO;
         } else {
           fdc_fifo_locked = 1;
+          result = 0;
         }
       }
     }
+  } else {
+    TRACE( 2, "Fifo not enabled because locked");
   }
   TRACE_EXIT;
   return result;
 }
 
+/*   Determine fd controller type 
+ *
+ *
+ */
+
+static byte fdc_save_state[ 2] = { 0, 0 };
+
 int
 fdc_probe( void)
 {
   TRACE_FUN( 8, "fdc_probe");
-  byte cmd = FDC_VERSION;
-  byte stat;
+  byte cmd[ 1];
+  byte stat[ 10];       /* must be able to hold dumpregs results */
   int result;
 
   /*  Try to find out what kind of fd controller we have to deal with
+   *  Scheme borrowed from floppy driver:
+   *  first try if FDC_DUMPREGS command works
+   *  (this indicates that we have a 82072 or better)
+   *  then try the FDC_VERSION command (82072 doesn't support this)
+   *  then try the FDC_UNLOCK command (some older 82077's don't support this)
    */
-  result = fdc_issue_command( &cmd, 1, &stat, 1);
-  if (result < 0) {
-    TRACE( -1, "Couldn't determine fdc version, assume 8272");
-    result = i8272;             /* safe choice */
-  } else {
-    if (stat == 0x90) {
-      TRACE( 2, "Type 82077 or compatible FDC found");
-      result = i82077;
-    } else if (stat == 0x80) {
-      TRACE( 2, "Type 8272A or compatible FDC found");
+  cmd[ 0] = FDC_DUMPREGS;
+  result = fdc_issue_command( cmd, 1, stat, 1);
+  if (result == 0) {
+    if (stat[ 0] == 0x80) {
+      /* invalid command: must be pre 82072
+       */
+      TRACE( 2, "Type 8272A/765A compatible FDC found");
       result = i8272;
     } else {
-      TRACEx1( -1, "Unknown FDC type found: 0x%02x", stat);
-      result = i8272;
+      fdc_result( &stat[ 1], 9);
+      fdc_save_state[ 0] = stat[ 7];
+      fdc_save_state[ 1] = stat[ 8];
+      cmd[ 0] = FDC_VERSION;
+      result = fdc_issue_command( cmd, 1, stat, 1);
+      if (result < 0 || stat[ 0] == 0x80) {
+        TRACE( 2, "Type 82072 FDC found");
+        result = i8272;
+      } else if (*stat == 0x90) {
+        cmd[ 0] = FDC_UNLOCK;
+        result = fdc_issue_command( cmd, 1, stat, 1);
+        if (result < 0 || stat[ 0] != 0x00) {
+          TRACE( 2, "Type pre-1991 82077 FDC found, treating it like a 82072");
+          result = i8272;
+        } else {
+          int i;
+
+          if (fdc_save_state[ 0] & 0x80) { /* was locked */
+            cmd[ 0] = FDC_LOCK; /* restore lock */
+            result = fdc_issue_command( cmd, 1, stat, 1);
+            TRACE( 2, "FDC is already locked");
+          }
+          result = no_fdc;
+          for (i = 0 ; i < 4 ; ++i) {
+            outb_p( i, fdc.tdr);
+            if ((inb_p( fdc.tdr) & 0x03) != i) {
+              result = i82077;
+              break;
+            }
+          }
+          if (result == no_fdc) {
+            result = i82077AA;
+            TRACE( 2, "Type 82077AA FDC found");
+          } else {
+            TRACE( 2, "Type 82077 FDC found");
+          }
+        }
+      } else {
+        TRACE( 2, "Unknown FDC found");
+        result = i8272;
+      }
     }
+  } else {
+    TRACE( -1, "No FDC found");
+    result = no_fdc;
   }
   TRACE_EXIT;
   return result;
@@ -839,12 +948,12 @@ fdc_config_regs( unsigned fdc_base, unsigned fdc_irq, unsigned fdc_dma)
   fdc.tdr = fdc_base + 3;
   fdc.msr = fdc.dsr = fdc_base + 4;
   fdc.fifo = fdc_base + 5;
+#ifdef MACH2
+  fdc.dor2 = fdc_base + 6;
+#endif
   fdc.dir = fdc.ccr = fdc_base + 7;
 }
 
-/*  The code that follows will be a lot cleaner when the
- *  irq and dma can be shared with the floppy driver.
- */
 /*  If probing for a FC-10 controller the fdc base address, interrupt
  *  and dma channel must be specified.
  *  If using an alternate fdc controller, base address, interrupt and
@@ -875,12 +984,13 @@ fdc_config( void)
       fdc.hook = &do_ftape;
     } else {
       TRACE( 2, "FC-10 controller not found");
+      fdc.type = no_fdc;
       fdc_config_regs( 0x3f0, 6, 2); /* back to std fdc again */
-#ifdef FLOPPY_HACK
+# ifdef FLOPPY_HACK
       fdc.hook = &do_floppy;
-#else
+# else
       fdc.hook = &do_ftape;
-#endif
+# endif
     }
 #else
 # ifdef FDC_BASE
@@ -890,11 +1000,11 @@ fdc_config( void)
 # else
     TRACE( 2, "Using the standard fdc controller");
     fdc_config_regs( 0x3f0, 6, 2); /* std fdc */
-#ifdef FLOPPY_HACK
+#  ifdef FLOPPY_HACK
     fdc.hook = &do_floppy;
-#else
+#  else
     fdc.hook = &do_ftape;
-#endif
+#  endif
 # endif
 #endif
   }
@@ -903,21 +1013,7 @@ fdc_config( void)
   TRACE_EXIT;
 }
 
-#ifdef FLOPPY_HACK
-
-inline int
-fdc_grab_irq_and_dma( void)
-{
-  return 0;
-}
-
-inline int
-fdc_release_irq_and_dma( void)
-{
-  return 0;
-}
-
-#else
+#ifndef FLOPPY_HACK
 
 static void ftape_interrupt( int unused)
 {
@@ -933,25 +1029,36 @@ static void ftape_interrupt( int unused)
   TRACE_EXIT;
 }
 
+#endif
+
 int
 fdc_grab_irq_and_dma( void)
 {
   TRACE_FUN( 8, "fdc_grab_irq_and_dma");
   int result = 0;
-  struct sigaction ftape_sig_action = {
-    ftape_interrupt, 0, SA_INTERRUPT, NULL
-  };
+#ifndef FLOPPY_HACK
+  static char ftape_id[] = "ftape";
 
-  if (fdc.hook != &do_floppy) {
-    /*  Cannot use request_irq because we want a fast interrupt
-     *  handler instead of a normal one (see kernel/irq.c).
+  if (fdc.hook == &do_ftape) {
+    /*  Get fast interrupt handler.
      */
+#ifdef OLD_IRQ_REQUEST
+    struct sigaction ftape_sig_action = {
+      ftape_interrupt, 0, SA_INTERRUPT, NULL
+    };
     result = irqaction( fdc.irq, &ftape_sig_action);
+#else
+    result = request_irq( fdc.irq, ftape_interrupt, SA_INTERRUPT, ftape_id);
+#endif
     if (result) {
       TRACEx1( -1, "Unable to grab IRQ%d for ftape driver", fdc.irq);
       result = -EIO;
     } else {
+#ifdef OLD_DMA_REQUEST
       result = request_dma( fdc.dma);
+#else
+      result = request_dma( fdc.dma, ftape_id);
+#endif
       if (result) {
         TRACEx1( -1, "Unable to grab DMA%d for ftape driver", fdc.dma);
         free_irq( fdc.irq);
@@ -961,6 +1068,18 @@ fdc_grab_irq_and_dma( void)
       }
     }
   }
+#endif
+#ifdef FDC_DMA
+  if (result == 0 && FDC_DMA == 2) {
+    /*  Using same dma channel as standard fdc, need to disable the
+     *  dma-gate on the std fdc. This couldn't be done in the floppy
+     *  driver as some laptops are using the dma-gate to enter a
+     *  low power or even suspended state :-(
+     */
+    outb_p( FDC_RESET_NOT, 0x3f2);
+    TRACE( 2, "DMA-gate on standard fdc disabled");
+  }
+#endif
   TRACE_EXIT;
   return result;
 }
@@ -968,19 +1087,36 @@ fdc_grab_irq_and_dma( void)
 int
 fdc_release_irq_and_dma( void)
 {
+  TRACE_FUN( 8, "fdc_grab_irq_and_dma");
   int result = 0;
 
-  if (fdc.hook != &do_floppy) {
+#ifdef FLOPPY_HACK
+  if (fdc.hook == &do_floppy) {
+    do_floppy = NULL;           /* prevent accidents ! */
+  }
+#else
+  if (fdc.hook == &do_ftape) {
     disable_dma( fdc.dma);      /* just in case... */
     free_dma( fdc.dma);
     disable_irq( fdc.irq);
     free_irq( fdc.irq);
   }
+#endif
+#ifdef FDC_DMA
+  if (result == 0 && FDC_DMA == 2) {
+    /*  Using same dma channel as standard fdc, need to disable the
+     *  dma-gate on the std fdc. This couldn't be done in the floppy
+     *  driver as some laptops are using the dma-gate to enter a
+     *  low power or even suspended state :-(
+     */
+    outb_p( FDC_RESET_NOT| FDC_DMA_MODE, 0x3f2);
+    TRACE( 2, "DMA-gate on standard fdc enabled again");
+  }
+#endif
+  TRACE_EXIT;
   return result;
 }
 
-#endif
-
 int
 fdc_init( void)
 {
@@ -995,29 +1131,21 @@ fdc_init( void)
     fdc_catch_stray_interrupts( 1); /* one always comes */
     TRACE( 5, "resetting fdc");
     fdc_reset();                  /* init fdc & clear track counters */
-    if (fdc.type == i8272) {      /* default, means no fc-10 found */
+    if (fdc.type == no_fdc) {     /* default, means no fc-10 found */
       fdc.type = fdc_probe();
     }
-    if (fdc.type == i82077 || fdc.type == fc10) {
-      if (fdc_fifo_enable() >= 0) {
-        TRACE( 2, "FDC fifo successfully enabled");
+    if (fdc.type != no_fdc) {
+      if (fdc.type >= i82077) {
+        if (fdc_fifo_enable() < 0) {
+          TRACE( 2, "couldn't enable fdc fifo !");
+        } else {
+          TRACE( 2, "fdc fifo enabled and locked");
+        }
       }
-    }
-#ifdef SOFT_RESET
-    /*  The soft reset caused by the fdc_recalibrate command will always
-     *  set the new-cartridge status on Iomega make drives. This is probably
-     *  due to an QIC-117 spec interpretation error. Since we have to live
-     *  with it we'll try to prevent soft resets as much as possible.
-     *  The function of this call was to clear the track counter in the fdc
-     *  to zero. That might be done by the fdc_reset too (I'm not sure).
-     */
-    TRACE( 5, "recalibrating");
-    if (fdc_recalibrate() != 0) {
-      TRACE( 1, "recalibrate failed");
+    } else {
       fdc_release_irq_and_dma();
       result = -EIO;
     }
-#endif
   }
   TRACE_EXIT;
   return result;
