@@ -15,18 +15,18 @@ You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
- $Source: /usr/src/distr/ftape-1.13b/RCS/fdc-isr.c,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/fdc-isr.c,v $
  $Author: bas $
  *
- $Revision: 1.15 $
- $Date: 1994/07/05 20:18:02 $
+ $Revision: 1.30 $
+ $Date: 1995/01/15 15:56:46 $
  $State: Beta $
  *
  *      This file contains the interrupt service routine and associated
  *      code for the QIC-40/80 tape streamer device driver.
  */
 
-static char RCSid[] = "$Id: fdc-isr.c,v 1.15 1994/07/05 20:18:02 bas Beta bas $";
+static char RCSid[] = "$Id: fdc-isr.c,v 1.30 1995/01/15 15:56:46 bas Beta bas $";
 
 #include <asm/io.h>
 #include <asm/dma.h>
@@ -37,9 +37,10 @@ static char RCSid[] = "$Id: fdc-isr.c,v 1.15 1994/07/05 20:18:02 bas Beta bas $"
 #include "fdc-isr.h"
 #include "qic117.h"
 #include "fdc-io.h"
-#include "ftape-io.h"
+#include "ftape-ctl.h"
 #include "ftape-rw.h"
 #include "calibr.h"
+#include "ftape-bsm.h"
 
 /*      Global vars.
  */
@@ -47,6 +48,7 @@ volatile int expected_stray_interrupts = 0;
 volatile int seek_completed = 0;
 volatile int interrupt_seen = 0;
 volatile int expect_stray_interrupt = 0;
+int random_rw = 0;
 
 /*      Local vars.
  */
@@ -54,44 +56,69 @@ typedef enum { no_error = 0, id_am_error = 0x01, id_crc_error = 0x02,
                data_am_error = 0x04, data_crc_error = 0x08,
                no_data_error =0x10, overrun_error = 0x20,
              } error_cause;
-int hide_interrupt;
+static int hide_interrupt;
+static int stop_read_ahead = 0;
 
 
+static void
+print_error_cause( int cause)
+{
+  TRACE_FUN( 8, "print_error_cause");
+
+  switch (cause) {
+  case no_data_error:
+    TRACE( 4, "no data error");
+    break;
+  case id_am_error:
+    TRACE( 4, "id am error");
+    break;
+  case id_crc_error:
+    TRACE( 4, "id crc error");
+    break;
+  case data_am_error:
+    TRACE( 4, "data am error");
+    break;
+  case data_crc_error:
+    TRACE( 4, "data crc error");
+    break;
+  case overrun_error:
+    TRACE( 4, "overrun error");
+    break;
+  default:
+  }
+  TRACE_EXIT;
+}
+
+static char*
+get_fdc_mode_text( fdc_mode_enum fdc_mode)
+{
+  switch (fdc_mode) {
+  case fdc_idle:
+    return "fdc_idle";
+  case fdc_reading_data:
+    return "fdc_reading_data";
+  case fdc_seeking:
+    return "fdc_seeking";
+  case fdc_writing_data:
+    return "fdc_writing_data";
+  case fdc_reading_id:
+    return "fdc_reading_id";
+  case fdc_recalibrating:
+    return "fdc_recalibrating";
+  default:
+    return "unknown";
+  }
+}
+
 static void
 decode_irq_cause( fdc_mode_enum fdc_mode, byte st[],
                  char** fdc_mode_txt, error_cause* cause)
 {
   TRACE_FUN( 8, "decode_irq_cause");
-  char* fdc_mode_text[] = {
-    "fdc_idle", "fdc_reading_data", "fdc_seeking",
-    "fdc_writing_data", "fdc_reading_id", "fdc_recalibrating",
-  };
 
-  /*  Valid in, decode cause of interrupt.
+  /*  Valid st[], decode cause of interrupt.
    */
-  switch (fdc_mode) {
-  case fdc_idle:
-    *fdc_mode_txt = fdc_mode_text[ 0];
-    break;
-  case fdc_reading_data:
-    *fdc_mode_txt = fdc_mode_text[ 1];
-    break;
-  case fdc_seeking:
-    *fdc_mode_txt = fdc_mode_text[ 2];
-    break;
-  case fdc_writing_data:
-    *fdc_mode_txt = fdc_mode_text[ 3];
-    break;
-  case fdc_reading_id:
-    *fdc_mode_txt = fdc_mode_text[ 4];
-    break;
-  case fdc_recalibrating:
-    *fdc_mode_txt = fdc_mode_text[ 5];
-    break;
-  default:
-    *fdc_mode_txt = "unknown";
-    break;
-  }
+  *fdc_mode_txt = get_fdc_mode_text( fdc_mode);
   switch (st[ 0] & ST0_INT_MASK) {
   case FDC_INT_NORMAL:
     TRACEx1( 5, "normal completion: %s", *fdc_mode_txt);
@@ -120,6 +147,7 @@ decode_irq_cause( fdc_mode_enum fdc_mode, byte st[],
     } else if (st[ 1] & 0x10) {
       *cause = overrun_error;
     }
+    print_error_cause( *cause);
     break;
   case FDC_INT_INVALID:
     TRACEx1( 5, "invalid completion %s", *fdc_mode_txt);
@@ -161,62 +189,115 @@ update_history( error_cause cause)
 }
 
 static void
-determine_progress( buffer_struct* buff, error_cause cause)
+skip_bad_sector( buffer_struct* buff)
+{
+  TRACE_FUN( 8, "skip_bad_sector");
+
+  /* Mark sector as soft error and skip it
+   */
+  if (buff->remaining > 0) {
+    ++buff->sector_offset;
+    ++buff->data_offset;
+    --buff->remaining;
+    buff->ptr += SECTOR_SIZE;
+    buff->bad_sector_map >>= 1;
+  } else {
+    ++buff->sector_offset;      /* hack for error maps */
+    TRACE( 1, "skipping last sector in segment");
+  }
+  TRACE_EXIT;
+}
+
+static void
+update_error_maps( buffer_struct* buff, unsigned error_offset)
+{
+  TRACE_FUN( 8, "update_error_maps");
+  int hard = 0;
+
+  /* error_offset is a sector offset !
+   */
+  if (buff->retry < SOFT_RETRIES) {
+    buff->soft_error_map |= (1 << error_offset);
+  } else {
+    buff->hard_error_map |= (1 << error_offset);
+    buff->soft_error_map &= ~buff->hard_error_map;
+    buff->retry = -1;           /* will be set to 0 in setup_segment */
+    hard = 1;
+  }
+  TRACEx2( 4, "sector %d : %s error", SECTOR( error_offset),
+          hard ? "hard" : "soft");
+  TRACEx2( 5, "hard map: 0x%08lx, soft map: 0x%08lx",
+          buff->hard_error_map, buff->soft_error_map);
+  TRACE_EXIT;
+}
+
+/*
+ *  Error cause:   Amount xferred:  Action:
+ *
+ *  id_am_error         0           mark bad and skip
+ *  id_crc_error        0           mark bad and skip
+ *  data_am_error       0           mark bad and skip
+ *  data_crc_error    % 1024        mark bad and skip
+ *  no_data_error       0           retry on write
+ *                                  mark bad and skip on read
+ *  overrun_error  [ 0..all-1 ]     mark bad and skip
+ *  no_error           all          continue
+ */
+static void
+determine_progress( buffer_struct* buff, error_cause cause, int mode)
 {
   TRACE_FUN( 8, "determine_progress");
   unsigned nr_not_xferred;
+  unsigned nr_xferred;
   unsigned dma_residue;
 
   disable_dma( fdc.dma);
   dma_residue = get_dma_residue( fdc.dma);
-  if (cause != no_error || dma_residue > 0) {
+  nr_xferred = buff->sector_count * SECTOR_SIZE - dma_residue;
+  if (cause == no_error && dma_residue == 0) {
+    nr_not_xferred = 0;
+  } else {
     if (cause == no_error) {
-      TRACEx1( 6, "unexpected DMA residue: 0x%04x", dma_residue);
+      TRACEx1( 4, "unexpected DMA residue: 0x%04x", dma_residue);
     } else {
       TRACEx1( 6, "DMA residue = 0x%04x", dma_residue);
     }
     nr_not_xferred = ((dma_residue + (SECTOR_SIZE - 1)) / SECTOR_SIZE);
-  } else {
-    nr_not_xferred = 0;
-  }
-  if (cause != no_error && cause != data_crc_error &&
-      cause != overrun_error && cause != no_data_error) {
-    if (nr_not_xferred > 0) {
-      --nr_not_xferred;         /* skip error in sector not yet done */
-    } else {
-      TRACE( -1, "this is probably a bug :-(");
-    }
-  }
-  if (nr_not_xferred > 0) {     /* compensate for sectors not transferred */
-    buff->sector_count -= nr_not_xferred;
-    /* bad_sector_map is already adjusted for the situation if there
-     * were sector_count sectors done without error (calc_next_cluster()).
-     */
-    buff->bad_sector_map <<= nr_not_xferred; /* re-adjust */
+    buff->sector_count -= nr_not_xferred; /* adjust to actual value */
   }
   /*  Update var's influenced by the DMA operation.
    */
-  if (buff->sector_count != 0) {
+  if (buff->sector_count > 0) {
     buff->sector_offset += buff->sector_count;
     buff->data_offset += buff->sector_count;
     buff->ptr += buff->sector_count * SECTOR_SIZE;
     buff->remaining -= buff->sector_count;
+    buff->bad_sector_map >>= buff->sector_count;
   }
-  if (cause != no_error) {
-    if (cause != no_data_error) {
-      /* merge position of bad sector in bitmap
-       */
-      if (buff->data_offset > 0) {
-        buff->crc_error_map |=  (1 << (buff->data_offset - 1));
-        TRACEx1( 4, "Error at sector offset %d", buff->data_offset - 1);
-      } else {
-        TRACEx1( 4, "Unexpected error at sector offset %d", buff->data_offset);
-      }
-    } else {
-      TRACEx1( 4, "Sector at offset %d not found", buff->data_offset);
-    }
+  if (cause == no_error) {
+    TRACEx1( 5, "%d Sector(s) transferred", buff->sector_count);
+  } else if (cause == no_data_error) {
+    TRACEx1( 5, "Sector %d not found", SECTOR( buff->sector_offset));
+  } else if (nr_xferred > 0) {
+    TRACEx1( 5, "Error in sector %d", SECTOR( buff->sector_offset));
+  } else if (cause == overrun_error) {
+    /* got an overrun error on the first byte, must be a hardware problem
+     */
+    TRACE( -1, "Unexpected error: failing DMA controller ?");
   } else {
-    TRACEx1( 5, "%d Sectors successfully transferred", buff->sector_count);
+    TRACEx1( 4, "Unexpected error at sector %d", SECTOR( buff->sector_offset));
+  }
+  /*  Sector_offset points to the problem area, except if we got
+   *  a data_crc_error. In that case it points one past the failing
+   *  sector.
+   *  Now adjust sector_offset so it allways points one past he
+   *  failing sector. I.e. skip the bad sector.
+   */
+  if (cause != no_error) {
+    if (cause != data_crc_error) {
+      skip_bad_sector( buff);
+    }
+    update_error_maps( buff, buff->sector_offset - 1);
   }
   TRACE_EXIT;
 }
@@ -232,11 +313,12 @@ calc_steps( int cmd)
 }
 
 static void
-pause_tape( unsigned segment)
+pause_tape( unsigned segment, int retry)
 {
   TRACE_FUN( 8, "pause_tape");
   int result;
-  byte out[ 3] = { FDC_SEEK, FTAPE_UNIT, calc_steps( QIC_PAUSE)};
+  byte out[ 3] = { FDC_SEEK, FTAPE_UNIT,
+                   calc_steps( retry ? QIC_MICRO_STEP_PAUSE : QIC_PAUSE)};
   
   /*  We'll use a raw seek command to get the tape to rewind
    *  and stop for a retry.
@@ -244,8 +326,10 @@ pause_tape( unsigned segment)
   ++history.rewinds;
   result = fdc_command( out, 3); /* issue QIC_117 command */
   if (result < 0) {
-    TRACEx1( 5, "qic-pause failed, status = %d", result);
+    TRACEx1( 4, "qic-pause failed, status = %d", result);
   }
+  location.known = 0;
+  tape_running = 0;
   runner_status = idle;
   hide_interrupt = 1;
   TRACE_EXIT;
@@ -262,135 +346,173 @@ stop_tape( unsigned segment)
    */
   result = fdc_command( out, 3); /* issue QIC_117 command */
   if (result < 0) {
-    TRACEx1( 5, "qic-stop failed, status = %d", result);
+    TRACEx1( 4, "qic-stop failed, status = %d", result);
   } else {
     runner_status = idle;
     hide_interrupt = 1;
-    /*  Can't say where we are if at end of a track, start_tape will find out.
-     */
-    if ((segment + 1) % segments_per_track != 0) { /* not last on track */
-      current_segment = ++segment;  /* better too high than too low */
-      location_known = 1;
-      TRACEx1( 5, "stopping at segment = %d", segment);
-    } else {
-      TRACEx1( 5, "stopping at end of track %d", segment / segments_per_track);
-    }
+    tape_running = 0;
   }
   TRACE_EXIT;
 }
 
 static void
 continue_xfer( buffer_struct** p_buff, error_cause cause,
-              int write, unsigned skip)
+              int fdc_mode, unsigned skip)
 {
   TRACE_FUN( 8, "continue_xfer");
   buffer_struct* buff = *p_buff;
+  int write = (fdc_mode == fdc_writing_data);
   byte fdc_op = (write) ? FDC_WRITE : FDC_READ;
 
-  if (!skip) {
-    if (runner_status != running || (write && buff->status != writing) ||
-        (!write && buff->status != reading)) {
+  if (skip > 0) {
+    /* This part can be removed if it never happens
+     */
+    if (runner_status != running ||
+        (buff->status != (write ? writing : reading))) {
       TRACEx2( 1, "unexpected runner/buffer state %d/%d",
               runner_status, buff->status);
       buff->status = error;
       *p_buff = next_buffer( &head); /* finish this buffer */
       runner_status = aborting;
       fdc_mode = fdc_idle;
-      TRACE_EXIT;
-      return;                   /* dunno anything better to do */
     }
   }
   if (buff->remaining > 0 && calc_next_cluster( &buffer[ head]) > 0) {
-    /* next sector(s) */
+    /* still sectors left in current segment, continue with this segment
+     */
     if (setup_fdc_and_dma( &buffer[ head], fdc_op) < 0) {
+      /* failed, abort operation
+       */
       buff->bytes = buff->ptr - buff->address;
       buff->status = error;
       buff = *p_buff = next_buffer( &head); /* finish this buffer */
       runner_status = aborting;
       fdc_mode = fdc_idle;
     }
-  } else {                      /* next segment */
+  } else {
+    /* current segment completed
+     */
     unsigned last_segment = buff->segment_id;
     int eot = ((last_segment + 1) % segments_per_track) == 0;
     int next = buff->next_segment; /* 0 means stop ! */
 
     buff->bytes = buff->ptr - buff->address;
-    buff->status = (write) ? empty : full; /* segment done */
+    buff->status = done;
     buff = *p_buff = next_buffer( &head);
-    if (!eot) {                 /* read ahead / write next */
-      if (buff->status == ((write) ? full : empty) && (next > 0)) {
+    if (eot) {
+      /* finished last segment on current track, can't continue
+       */
+      runner_status = logical_eot;
+      fdc_mode = fdc_idle;
+    } else if (next > 0) {
+      /* continue with next segment
+       */
+      if (buff->status == waiting) {
         if (write && next != buff->segment_id) {
           TRACE( 5, "segments out of order, aborting write");
           runner_status = do_abort;
           fdc_mode = fdc_idle;
-          TRACE_EXIT;
-          return;
-        }
-        setup_new_segment( &buffer[ head], next, 0, SECTORS_PER_SEGMENT);
-        if (calc_next_cluster( &buffer[ head]) == 0 ||
-            setup_fdc_and_dma( &buffer[ head], fdc_op) != 0) {
-          TRACEx1( 1, "couldn't start %s-ahead", (write) ? "write" : "read");
-          runner_status = do_abort;
-          fdc_mode = fdc_idle;
         } else {
-          buff->status = (write) ? writing : reading; /* keep on going */
+          setup_new_segment( &buffer[ head], next, 0);
+          if (stop_read_ahead) {
+            buff->next_segment = 0;
+            stop_read_ahead = 0;
+          }
+          if (calc_next_cluster( &buffer[ head]) == 0 ||
+              setup_fdc_and_dma( &buffer[ head], fdc_op) != 0) {
+            TRACEx1( 1, "couldn't start %s-ahead", (write) ? "write" : "read");
+            runner_status = do_abort;
+            fdc_mode = fdc_idle;
+          } else {
+            buff->status = (write) ? writing : reading; /* keep on going */
+          }
         }
       } else {
-        if (next > 0) {
-          TRACEx1( 5, "all input buffers %s, pausing tape",
+        TRACEx1( 5, "all input buffers %s, pausing tape",
                 (write) ? "empty" : "full");
-          pause_tape( last_segment);
-        } else {
-          TRACEx1( 5, "no %s allowed, stopping tape",
-                (write) ? "write next" : "read ahead");
-          stop_tape( last_segment);
-        }
+        pause_tape( last_segment, 0);
         runner_status = idle;   /* not quite true until next irq */
       }
     } else {
-      runner_status = logical_eot; /* just got last segment on track */
-      fdc_mode = fdc_idle;
+      /* don't continue with next segment
+       */
+      TRACEx1( 5, "no %s allowed, stopping tape",
+              (write) ? "write next" : "read ahead");
+      if (random_rw) {
+        stop_tape( last_segment);
+      } else {
+        pause_tape( last_segment, 0);
+      }
+      runner_status = idle;     /* not quite true until next irq */
     }
   }
+  TRACE_EXIT;
+  return;
 }
 
 static void
 retry_sector( buffer_struct** p_buff, error_cause cause,
-             int write, unsigned skip)
+             int fdc_mode, unsigned skip)
 {
   TRACE_FUN( 8, "retry_sector");
   buffer_struct* buff = *p_buff;
 
-#if 0
-  if (buff->retry >= RETRIES_ON_CRC_ERROR) {
-    TRACE( 2, "too many retries, continue");
-    continue_xfer( p_buff, cause, write, 0);
-  } else {
-#endif
-    TRACEx1( 5, "%s error, preparing for retry", (write) ? "write" : "read");
-#if 1
-    runner_status = aborting;
-    fdc_mode = fdc_idle;
-#else
-    /*  problematic: hangs Linux in read_segment after wait_segment.
-     *  can ^C to exit :-)
-     */
-    pause_tape( buff->segment_id);
-    runner_status = idle;
-#endif
-    buff->status = error;
-    if (skip > 0) {
-      buff->skip = skip;
-      TRACEx1( 4, "request skipping of %d sectors", buff->skip);
-    } else {
-      buff->skip = 0;
-    }
-#if 0
-  }
-#endif
+  TRACEx1( 4, "%s error, will retry",
+          (fdc_mode == fdc_writing_data) ? "write" : "read");
+  pause_tape( buff->segment_id, 1);
+  runner_status = aborting;
+  buff->status = error;
+  buff->skip = skip;
   TRACE_EXIT;
 }
 
+static unsigned
+find_resume_point( buffer_struct* buff)
+{
+  TRACE_FUN( 8, "find_resume_point");
+  int i = 0;
+  unsigned long mask;
+  unsigned long map;
+
+  /*  This function is to be called after all variables have been
+   *  updated to point past the failing sector.
+   *  If there are any soft errors before the failing sector,
+   *  find the first soft error and return the sector offset.
+   *  Otherwise find the last hard error.
+   *  Note: there should allways be at least one hard or soft error !
+   */
+  if (buff->sector_offset < 1 || buff->sector_offset > 32) {
+    TRACEx1( 1, "bug: sector_offset = %d", buff->sector_offset);
+  } else {
+    if (buff->sector_offset >= 32) {
+      mask = 0xffffffff;
+    } else {
+      mask = (1 << buff->sector_offset) - 1;
+    }
+    map = buff->soft_error_map & mask;
+    if (map) {
+      while ((map & (1 << i)) == 0) {
+        ++i;
+      }
+      TRACEx1( 4, "at sector %d", SECTOR( i));
+    } else {
+      map = buff->hard_error_map & mask;
+      i = buff->sector_offset - 1;
+      if (map) {
+        while ((map & (1 << i)) == 0) {
+          --i;
+        }
+        TRACEx1( 4, "after sector %d", SECTOR( i));
+        ++i;                    /* first sector after last hard error */
+      } else {
+        TRACE( 1, "bug: no soft or hard errors");
+      }
+    }
+  }
+  TRACE_EXIT;
+  return i;
+}
+
 /*      FDC interrupt service routine.
  */
 void
@@ -402,8 +524,9 @@ fdc_isr( void)
   error_cause cause = no_error;
   byte in[ 7];
   static int isr_active = 0;
-  int t0, t1;
+  int t0;
   buffer_struct* buff = &buffer[ head];
+  int skip;
 
   t0 = timestamp();
   if (isr_active) {
@@ -437,7 +560,6 @@ fdc_isr( void)
       switch (fdc_mode) {
         
       case fdc_reading_data: {
-        
         TRACEi( 5, "fdc irq reading segment", buff->segment_id);
         if (runner_status == aborting || runner_status == do_abort) {
           TRACEx1( 5, "aborting %s", fdc_mode_txt);
@@ -446,7 +568,7 @@ fdc_isr( void)
         if (buff->retry > 0) {
           TRACEx1( 5, "this is retry nr %d", buff->retry);
         }
-        if (buff->bad_sector_map == ((unsigned) FAKE_SEGMENT) >> 1) {
+        if (buff->bad_sector_map == FAKE_SEGMENT) {
           /* This condition occurs when reading a `fake' sector that's
            * not accessible. Doesn't really matter as we would have
            * ignored it anyway ! Might be located at wrong segment,
@@ -454,90 +576,61 @@ fdc_isr( void)
            */
           TRACE( 5, "skipping empty segment (read)");
           buff->remaining = 0; /* skip failing sector */
-          continue_xfer( &buff, no_error, 0, 1); /* fake success */
+          continue_xfer( &buff, no_error, fdc_mode, 1); /* fake success */
         } else {
           switch (cause) {
           case no_error: {
-            determine_progress( buff, cause);
-            continue_xfer( &buff, cause, 0, 0);
+            determine_progress( buff, cause, fdc_reading_data);
+            if (in[ 2] & 0x40) {
+              /*  Handle deleted data in header segments.
+               *  Skip segment and force read-ahead.
+               */
+              TRACEx1( 2, "deleted data in sector %d",
+                      SECTOR( buff->sector_offset - 1));
+              buff->deleted = 1;
+              buff->remaining = 0; /* abort transfer */
+              buff->soft_error_map |= (-1L << buff->sector_offset);
+              if (buff->segment_id == 0) {
+                stop_read_ahead = 1; /* stop on next segment */
+              }
+              buff->next_segment = buff->segment_id + 1; /* force read-ahead */
+              skip = (SECTORS_PER_SEGMENT - buff->sector_offset);
+            } else {
+              skip = 0;
+            }
+            continue_xfer( &buff, cause, fdc_mode, skip);
             break;
           }
-          case no_data_error: {
-            int skip;
-
+          case no_data_error:
             /*  Tape started too far ahead of or behind the right sector.
              *  This may also happen in the middle of a segment !
+             *  Handle no-data as soft error. If next sector fails too,
+             *  a retry (with needed reposition) will follow.
              */
-            update_history( cause);
-            determine_progress( buff, cause);
-            if (buff->sector_count > 0) {
-              /*  Get here if at least one sector was read successfully
-               *  before some sector id couldn't be found.
-               */
-              TRACEx1( 5, "partial success: %d sectors done up to now",
-                      buff->sector_count);
-              skip = buff->sector_offset; /* retry error sector */
-            } else if (buff->data_offset > 0 &&
-                buff->crc_error_map & (1 << (buff->data_offset - 1))) {
-              /*  Get here if the sector id not found comes right
-               *  after a failing sector.
-               */
-              TRACE( 5, "error in previous sector: resume problem");
-              if (buff->data_offset > 1 &&
-                  buff->crc_error_map & ((1 << (buff->data_offset - 1)) - 1)) {
-                TRACE( 5, "not first error in this segment");
-                skip = buff->sector_offset - 1; /* retry error sector */
-              } else {
-                TRACE( 5, "first error in this segment");
-                skip = buff->sector_offset; /* skip error sector */
-              }
-            } else {
-              TRACE( 5, "no previous error: resume problem");
-              if (buff->data_offset == 0 && buff->retry > 1) {
-                ++buff->start_offset;
-                TRACE( 5, "incremented start offset");
-              }
-              skip = buff->sector_offset;
-            }
-            retry_sector( &buff, cause, 0, skip);
-            break;
-          }
           case id_am_error:
           case id_crc_error:
-          case data_am_error: {
-            int first_error = (buff->crc_error_map == 0);
-
-            update_history( cause);
-            determine_progress( buff, cause);
-            if (first_error) {
-              /*  Let ecc code correct this single error
-               */
-              TRACE( 4, "single am/crc error, let ecc handle it");
-              continue_xfer( &buff, cause, 0, 0);
-            } else {
-              /*  Retry to minimize risk of failing ecc code
-               */
-              TRACE( 4, "too many errors, must retry");
-              retry_sector( &buff, cause, 0, 0);
-            }
-            break;
-          }
+          case data_am_error:
           case data_crc_error:
           case overrun_error: {
-            int first_error = (buff->crc_error_map == 0);
-
-            determine_progress( buff, cause);
+            int first_error = (buff->soft_error_map == 0 &&
+                               buff->hard_error_map == 0);
+        
             update_history( cause);
+            determine_progress( buff, cause, fdc_reading_data);
             if (first_error) {
-              if (cause == data_crc_error) {
-                TRACE( 4, "single crc error in data");
-              } else {
-                TRACE( 4, "single overrun error");
-              }
-              continue_xfer( &buff, cause, 0, 0);
+              skip = buff->sector_offset;
             } else {
-              TRACE( 4, "too many errors, must retry");
-              retry_sector( &buff, cause, 0, 0);
+              skip = find_resume_point( buff);
+            }
+            /*  Try to resume with next sector on single errors (let ecc
+             *  correct it), but retry on no_data (we'll are past the
+             *  target when we get here so we cannot retry) or on multiple
+             *  errors (reduce chance on ecc failure).
+             */
+            if (first_error && cause != no_data_error) {
+              continue_xfer( &buff, cause, fdc_mode, skip);
+            } else {
+              retry_sector( &buff, cause, fdc_mode, skip);
             }
             break;
           }
@@ -545,8 +638,8 @@ fdc_isr( void)
             /*  Don't know why this could happen but find out.
              */
             TRACE( 1, "unexpected error");
-            determine_progress( buff, cause);
-            retry_sector( &buff, cause, 0, 0);
+            determine_progress( buff, cause, fdc_reading_data);
+            retry_sector( &buff, cause, fdc_mode, 0);
             break;
           }
           }
@@ -581,7 +674,7 @@ fdc_isr( void)
         if (buff->retry > 0) {
           TRACEx1( 5, "this is retry nr %d", buff->retry);
         }
-        if (buff->bad_sector_map == ((unsigned) FAKE_SEGMENT) >> 1) {
+        if (buff->bad_sector_map == FAKE_SEGMENT) {
           /* This condition occurs when trying to write to a `fake'
            * sector that's not accessible. Doesn't really matter as
            * it isn't used anyway ! Might be located at wrong segment,
@@ -589,49 +682,23 @@ fdc_isr( void)
            */
           TRACE( 5, "skipping empty segment (write)");
           buff->remaining = 0; /* skip failing sector */
-          continue_xfer( &buff, no_error, 1, 1); /* fake success */
+          continue_xfer( &buff, no_error, fdc_mode, 1); /* fake success */
         } else {
           switch (cause) {
           case no_error: {
-            determine_progress( buff, cause);
-            continue_xfer( &buff, cause, 1, 0);
+            determine_progress( buff, cause, fdc_writing_data);
+            continue_xfer( &buff, cause, fdc_mode, 0);
             break;
           }     
-          case no_data_error: {
-            /*  Tape started too far ahead of or behind the right sector.
-             *  This happens in the middle of a segment too !
-             */
-            update_history( cause);
-            determine_progress( buff, cause);
-            if (buff->sector_count > 0) {
-              TRACEx1( 5, "partial success: %d sectors done up to now",
-                      buff->sector_count);
-            } else if (buff->data_offset == 0) {
-              TRACE( 5, "sector not found: start problem");
-              if (buff->retry > 1) {
-                ++buff->start_offset;
-                TRACE( 5, "incremented start offset");
-              }
-            } else {
-              TRACE( 5, "sector not found: resume problem");
-            }
-            retry_sector( &buff, cause, 1, buff->sector_offset);
-            break;
-          }
+          case no_data_error:
           case id_am_error:
           case id_crc_error:
-          case data_am_error: {
-            update_history( cause);
-            TRACE( 4, "single am/crc error");
-            determine_progress( buff, cause);
-            retry_sector( &buff, cause, 1, buff->sector_offset);
-            break;
-          }
+          case data_am_error:
           case overrun_error: {
             update_history( cause);
-            TRACE( 4, "single overrun error");
-            determine_progress( buff, cause);
-            retry_sector( &buff, cause, 1, buff->sector_offset);
+            determine_progress( buff, cause, fdc_writing_data);
+            skip = find_resume_point( buff);
+            retry_sector( &buff, cause, fdc_mode, skip);
             break;
           }
           default: {
@@ -696,11 +763,7 @@ fdc_isr( void)
       wake_up_interruptible( &wait_intr);
     }
   }
-  t1 = timestamp();
-  if (t1 < t0) {
-    t1 += LATCH;
-  }
-  t0 = (10000 * (t1 - t0)) / (CLOCK_TICK_RATE / 100);
+  t0 = timediff( t0, timestamp());
   if (t0 >= 1000) {             /* only tell us about long calls */
     TRACEx1( 7, "isr() duration: %5d usec", t0);
   }
