@@ -15,11 +15,11 @@ You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
- $Source: /usr/src/distr/ftape-1.13a/RCS/ftape-read.c,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/ftape-read.c,v $
  $Author: bas $
  *
- $Revision: 1.9 $
- $Date: 1994/06/29 20:11:06 $
+ $Revision: 1.23 $
+ $Date: 1995/01/15 15:56:46 $
  $State: Beta $
  *
  *      This file contains the reading code
@@ -27,24 +27,23 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
 static char RCSid[] =
-"$Id: ftape-read.c,v 1.9 1994/06/29 20:11:06 bas Beta $";
+"$Id: ftape-read.c,v 1.23 1995/01/15 15:56:46 bas Beta bas $";
 
 
 #include <linux/string.h>
 #include <linux/errno.h>
-#include <linux/sched.h>
-#include <asm/dma.h>
 #include <asm/segment.h>
-#include <asm/system.h>
 
 #include "ftape.h"
 #include "ftape-read.h"
 #include "qic117.h"
 #include "ftape-io.h"
+#include "ftape-ctl.h"
 #include "ftape-rw.h"
 #include "ftape-write.h"
 #include "ftape-eof.h"
 #include "ecc.h"
+#include "ftape-bsm.h"
 
 /*      Global vars.
  */
@@ -55,6 +54,59 @@ int just_before_eof = 0;
 /*      Local vars.
  */
 
+void
+ftape_zap_read_buffers( void)
+{
+  int i;
+
+  for (i = 0; i < NR_ITEMS( buffer); ++i) {
+    buffer[ i].address = ((byte*) tape_buffer) + i * BUFF_SIZE;
+    buffer[ i].status = waiting;
+    buffer[ i].bytes = 0;
+    buffer[ i].skip = 0;
+    buffer[ i].retry = 0;
+  }
+  buf_len_rd = 0;
+  buf_pos_rd = 0;
+  eof_mark = 0;
+  ftape_state = idle;
+}
+
+static unsigned long
+convert_sector_map( buffer_struct* buff)
+{
+  TRACE_FUN( 8, "convert_sector_map");
+  int i = 0;
+  unsigned long bad_map = get_bad_sector_entry( buff->segment_id);
+  unsigned long src_map = buff->soft_error_map | buff->hard_error_map;
+  unsigned long dst_map = 0;
+
+  if (bad_map || src_map) {
+    TRACEx1( 5, "bad_map = 0x%08lx", bad_map);
+    TRACEx1( 5, "src_map = 0x%08lx", src_map);
+  }
+  while (bad_map) {
+    while ((bad_map & 1) == 0) {
+      if (src_map & 1) {
+        dst_map |= (1 << i);
+      }
+      src_map >>= 1;
+      bad_map >>= 1;
+      ++i;
+    }
+    /* (bad_map & 1) == 1 */
+    src_map >>= 1;
+    bad_map >>= 1;
+  }
+  if (src_map) {
+    dst_map |= (src_map << i);
+  }
+  if (dst_map) {
+    TRACEx1( 5, "dst_map = 0x%08lx", dst_map);
+  }
+  TRACE_EXIT;
+  return dst_map;
+}
 
 int
 correct_and_copy( unsigned int tail, byte* destination)
@@ -64,7 +116,7 @@ correct_and_copy( unsigned int tail, byte* destination)
   int result;
   BAD_SECTOR read_bad;
 
-  mseg.read_bad = buffer[ tail].crc_error_map;
+  mseg.read_bad = convert_sector_map( &buffer[ tail]);
   mseg.marked_bad = 0;          /* not used... */
   mseg.blocks = buffer[ tail].bytes / SECTOR_SIZE;
   mseg.data = buffer[ tail].address;
@@ -121,7 +173,7 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
     if (ftape_state == writing) {
       ftape_flush_buffers();    /* flush write buffer */
       TRACE( 5, "calling ftape_abort_operation");
-      result = ftape_abort_operation( &current_segment);
+      result = ftape_abort_operation();
       if (result < 0) {
         TRACE( 1, "ftape_abort_operation failed");
         TRACE_EXIT;
@@ -129,7 +181,7 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
       }
     } else {
       /* clear remaining read buffers */
-      ftape_zap_buffers();
+      ftape_zap_read_buffers();
     }
     ftape_state = reading;
   }
@@ -142,15 +194,24 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
     /*    Search all full buffers for the first matching the wanted segment.
      *    Clear other buffers on the fly.
      */
-    while (!read_done && buffer[ tail].status == full) {
+    while (!read_done && buffer[ tail].status == done) {
       if (buffer[ tail].segment_id == segment_id) {
         unsigned eof_sector;
         unsigned sector_count = 0;
-        unsigned long bsm = bad_sector_map[ segment_id];
+        unsigned long bsm = get_bad_sector_entry( segment_id);
         int i;
 
         /*        If out buffer is already full, return its contents.
          */
+        if (buffer[ tail].deleted) {
+          TRACEi( 5, "found segment in cache :", segment_id);
+          TRACE_EXIT;
+          /*  Return a value that read_header_segment understands.
+           *  As this should only occur when searching for the header
+           *  segments it shouldn't be misinterpreted elsewhere.
+           */
+          return 0;
+        }
         TRACEi( 5, "found segment in cache :", segment_id);
         eof_sector = check_for_eof( segment_id);
         if (eof_sector > 0) {
@@ -191,7 +252,7 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
       } else {
         TRACEi( 5, "zapping segment in cache :", buffer[ tail].segment_id);
       }
-      buffer[ tail].status = empty;
+      buffer[ tail].status = waiting;
       next_buffer( &tail);
     }
     if (!read_done && buffer[ tail].status == reading) {
@@ -203,24 +264,19 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
             return result;
           }
           TRACE( 1, "wait_segment failed while reading");
-#if 0
-          TRACE_EXIT;
-          return -EIO;
-#else
-          ftape_abort_operation( &current_segment);
-#endif
+          ftape_abort_operation();
         }
       } else {
         /*        We're reading the wrong segment, stop runner.
          */
-        ftape_abort_operation( &current_segment);
+        ftape_abort_operation();
       }
     }
     /*    if just passed the last segment on a track, wait for BOT or EOT mark.
      */
     if (runner_status == logical_eot) {
       int status;
-      result = ftape_ready_wait( 10 * SECOND, &status);
+      result = ftape_ready_wait( timeout.seek, &status);
       if (result < 0) {
         TRACE( 1, "ftape_ready_wait waiting for eot/bot failed");
       }
@@ -233,12 +289,16 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
      */
     if (runner_status == aborting || runner_status == buffer_overrun ||
         runner_status == end_of_tape) {
-      if (runner_status != end_of_tape) {
-        ftape_smart_stop( &location_known, &current_segment);
+      if (runner_status != end_of_tape &&
+          !(runner_status == aborting && !tape_running)) {
+        ftape_dumb_stop();
       }
       if (runner_status == aborting) {
         if (buffer[ head].status == reading || buffer[ head].status == error) {
-          buffer[ head].status = empty;
+          if (buffer[ head].status == error) {
+            history.defects += count_ones( buffer[ head].hard_error_map);
+          }
+          buffer[ head].status = waiting;
         }
       }
       runner_status = idle;       /* aborted ? */
@@ -246,7 +306,7 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
     /*    If segment to read is empty, do not start runner for it,
      *    but wait for next read call.
      */
-    if (bad_sector_map[ segment_id] == EMPTY_SEGMENT) {
+    if (get_bad_sector_entry( segment_id) == EMPTY_SEGMENT) {
       bytes_read = 0;           /* flag empty segment */
       read_done = 1;
     }
@@ -272,14 +332,15 @@ read_segment( unsigned segment_id, byte* address, int* eof_mark,
             head, tail, runner_status);
     TRACEx2( 8, "buffer[].status, [head]: %d, [tail]: %d",
             buffer[ head].status, buffer[ tail].status);
-    if (buffer[ tail].status == empty) {
-      setup_new_segment( &buffer[ head], segment_id, -1, SECTORS_PER_SEGMENT);
+    if (buffer[ tail].status == waiting) {
+      setup_new_segment( &buffer[ head], segment_id, -1);
       if (!read_ahead) {
         buffer[ head].next_segment = 0; /* disable read-ahead */
       }
       calc_next_cluster( &buffer[ head]);
       if (runner_status == idle) {
-        result= ftape_start_tape( segment_id);
+        result = ftape_start_tape( segment_id,
+                                  buffer[ head].sector_offset);
         if (result < 0) {
           TRACEx1( 1, "Error: segment %d unreachable", segment_id);
           TRACE_EXIT;
@@ -308,58 +369,62 @@ read_header_segment( byte* address)
   int i;
   int result;
   int header_segment = -1;
-  unsigned int tape_segments_per_track;
-  unsigned int tape_tracks_per_cartridge;
   unsigned int max_floppy_side;
   unsigned int max_floppy_track;
   unsigned int max_floppy_sector;
+  int first_failed = 0;
   int status;
 
   result = ftape_report_drive_status( &status);
-  if (!(status & QIC_STATUS_REFERENCED)) {
-    TRACE( 1, "error: tape not formatted");
+  if (result < 0) {
+    TRACE( 1, "error: error_status or report failure");
     TRACE_EXIT;
     return -EIO;
   }
-  /*    Find first header segment on tape.
-   *      THIS IS NOT TESTED !!! Can only be tested with
-   *      a tape that has bad sectors in first segment(s).
-   *    Get start of data area and verify tape parameters.
-   */
-  ftape_last_segment.id = 5;    /* will allow us to read the header ! */
   TRACE( 5, "reading...");
-  do {
-    ++header_segment;
-    bad_sector_map[ header_segment] = 0x00000000;
+  ftape_last_segment.id = 68;   /* will allow us to read the header ! */
+  /*  We're looking for the first header segment.
+   *  A header segment cannot contain bad sectors, therefor at the
+   *  tape start, segments with bad sectors are (according to QIC-40/80)
+   *  written with deleted data marks and must be skipped.
+   */
+  used_header_segment = -1;
+  result = 0;
+  for (header_segment = 0;
+       header_segment < ftape_last_segment.id && result == 0;
+       ++header_segment) {
+    /*  Set no read-ahead, the isr will force read-ahead whenever
+     *  it encounters deleted data !
+     */
     result = read_segment( header_segment, address, &status, 0);
-    if (header_segment > 5) {
-      /* probably something else wrong, don't keep stuck
-       * in this loop !
-       */
-      TRACE( 1, "header segment not found");
-      TRACE_EXIT;
-      return -EIO;
+    if (result < 0 && !first_failed) {
+      TRACE( 1, "header segment damaged, trying backup");
+      first_failed = 1;
+      result = 0;               /* force read of next (backup) segment */
     }
-  } while (result < 0);
-  result = ftape_abort_operation( &current_segment);
+  }
+  if (result < 0 || header_segment >= ftape_last_segment.id) {
+    TRACE( 1, "no readable segment found");
+    TRACE_EXIT;
+    return -EIO;
+  }
+  result = ftape_abort_operation();
   if (result < 0) {
     TRACE( 1, "ftape_abort_operation failed");
     TRACE_EXIT;
     return -EIO;
   }
   if (GET4( address, 0) != 0xaa55aa55) {
-    TRACE( 1, "tape not formatted (wrong signature)");
+    TRACE( 1, "wrong signature in header segment");
     TRACE_EXIT;
     return -EIO;
   }
   header_segment_1 = GET2( address, 6);
   header_segment_2 = GET2( address, 8);
-  if (header_segment != header_segment_1 && /* first header */
-      header_segment != header_segment_2) { /* duplicate */
-    TRACE( 1, "wrong segment nr");
-    TRACE_EXIT;
-    return -EIO;
-  }
+  TRACEx2( 2, "header segments are %d and %d",
+          header_segment_1, header_segment_2);
+  used_header_segment = (first_failed) ? header_segment_2 : header_segment_1;
+
   /*    Verify tape parameters...
    *    QIC-40/80 spec:                 tape_parameters:
    *
@@ -373,87 +438,90 @@ read_header_segment( byte* address)
    *    max-floppy-sector               segments_per_cylinder *
    *                                    SECTORS_PER_SEGMENT
    */
-
-  tape_segments_per_track = GET2( address, 24);
-  tape_tracks_per_cartridge = *(address + 26);
+  format_code = (format_type) *(address + 4);
+  segments_per_track = GET2( address, 24);
+  tracks_per_tape = *(address + 26);
   max_floppy_side = *(address + 27);
   max_floppy_track = *(address + 28);
   max_floppy_sector = *(address + 29);
-
-  TRACEi( 4, "segments-per-track:  ", tape_segments_per_track);
-  TRACEi( 4, "tracks-per-cartridge:", tape_tracks_per_cartridge);
-  TRACEi( 4, "max-floppy-side:     ", max_floppy_side);
-  TRACEi( 4, "max-floppy-track:    ", max_floppy_track);
-  TRACEi( 4, "max-floppy-sector:   ", max_floppy_sector);
-
-  if (tape_segments_per_track == 0 && tape_tracks_per_cartridge == 0 &&
+  TRACEx6( 4, "(fmt/spt/tpc/fhm/ftm/fsm) = %d/%d/%d/%d/%d/%d",
+          format_code, segments_per_track, tracks_per_tape,
+          max_floppy_side, max_floppy_track, max_floppy_sector);
+  switch (format_code) {
+  case 2:
+  case 5:
+    break;
+  case 3:
+    TRACE( 1, "1100 ft tape : experimental support !");
+    break;
+  case 4:
+    TRACE( 1, "variable length (wide) tape : experimental support !");
+    break;
+  default:
+    TRACE( 1, "unknown tape format, please report !");
+    TRACE_EXIT;
+    return -EIO;
+  }    
+  if (segments_per_track == 0 && tracks_per_tape == 0 &&
       max_floppy_side == 0 && max_floppy_track == 0 &&
       max_floppy_sector == 0) {
-    /*
-     *  QIC-40 Rev E and earlier
+    /*  QIC-40 Rev E and earlier has no values in the header.
      */
-    tape_segments_per_track = 68; 
-    tape_tracks_per_cartridge = 20;
+    segments_per_track = 68; 
+    tracks_per_tape = 20;
     max_floppy_side = 1;
     max_floppy_track = 169;
     max_floppy_sector = 128;
   }
-
-#ifdef CONNER_BUG
-  /*  If included, this test will compensate for the wrong parameter
-   *  on tapes formatted by Conner's DOS software.
-   *  Maybe there are other bad combinations, but this is the only
-   *  one I've heard of up to now - 03/01/94 Bas.
+  /*  This test will compensate for the wrong parameter on tapes
+   *  formatted by Conner software.
    */
-  if (tape_segments_per_track == 150 &&
-      tape_tracks_per_cartridge == 28 &&
+  if (segments_per_track == 150 &&
+      tracks_per_tape == 28 &&
       max_floppy_side == 7 &&
       max_floppy_track == 149 &&
       max_floppy_sector == 128) {
+    TRACE( -1, "the famous CONNER bug: max_floppy_side off by one !");
     max_floppy_side = 6;
   }
-#endif
-  /*
-   *    in case report_drive_configuration command failed,
-   *    we set the parameters according to the tape data.
+  /*  This test will compensate for the wrong parameter on tapes
+   *  formatted by Colorado Windows software.
    */
-  if (unknown_drive_config) {
-    segments_per_track = tape_segments_per_track;
-    tracks_per_tape = tape_tracks_per_cartridge;
-    segments_per_head = ( (tracks_per_tape + max_floppy_side) /
-                         (max_floppy_side + 1) ) * segments_per_track;
+  if (segments_per_track == 150 &&
+      tracks_per_tape == 28 &&
+      max_floppy_side == 6 &&
+      max_floppy_track == 150 &&
+      max_floppy_sector == 128) {
+    TRACE( -1, "the famous Colorado bug: max_floppy_track off by one !");
+    max_floppy_track = 149;
   }
+  segments_per_head = ((max_floppy_sector / SECTORS_PER_SEGMENT) *
+                       (max_floppy_track + 1));
   /*
    *    Verify drive_configuration with tape parameters
    */
-  if ( (segments_per_track != tape_segments_per_track) ||
-      (tracks_per_tape != tape_tracks_per_cartridge) ||
+  if ( segments_per_head == 0 || segments_per_cylinder == 0 ||
       ((segments_per_track * tracks_per_tape - 1) / segments_per_head
        != max_floppy_side) ||
       (segments_per_head / segments_per_cylinder - 1 != max_floppy_track) ||
-      (segments_per_cylinder * SECTORS_PER_SEGMENT != max_floppy_sector) ) {
-    TRACE( 1, "Tape parameters error");
+      (segments_per_cylinder * SECTORS_PER_SEGMENT != max_floppy_sector)
+#ifndef TESTING
+      || (format_code == 4 && (max_floppy_track != 254 || max_floppy_sector != 128))
+#endif
+      ) {
+    TRACE( 1, "Tape parameters inconsistency, please report");
     TRACE_EXIT;
     return -EIO;
   }
   first_data_segment = GET2( address, 10); /* first data segment */
   TRACEi( 4, "first data segment:", first_data_segment);
-  /* Copy the bad sector map into our local buffer.
-   */
-  memcpy( bad_sector_map, address + 2 * SECTOR_SIZE, sizeof( bad_sector_map));
-#if 0
-  /* for testing of bad sector handling at end of tape
-   */
-  bad_sector_map[ segments_per_track * tracks_per_tape - 3] = 0x000003e0;
-  bad_sector_map[ segments_per_track * tracks_per_tape - 2] = 0xff3fffff;
-  bad_sector_map[ segments_per_track * tracks_per_tape - 1] = 0xffffe000;
-#endif
+  extract_bad_sector_map( address);
   /*  Find the highest segment id that allows still one full
    *  deblock_buffer to be written to tape.
    */
   ftape_last_segment.size = 0;
   for (i = segments_per_track * tracks_per_tape - 1; i >= 0; --i) {
-    int space = SECTORS_PER_SEGMENT - 3 - count_ones( bad_sector_map[ i]);
+    int space = SECTORS_PER_SEGMENT - 3 - count_ones( get_bad_sector_entry( i));
     if (space > 0) {
       ftape_last_segment.size += space; /* sectors free */
       ftape_last_segment.free = (ftape_last_segment.size -
@@ -466,70 +534,14 @@ read_header_segment( byte* address)
       }
     }
   }
-#if 0
-  /*  Enable to test error recovery on Verbatime tape
-   *  with bad area.
-   */
-  bad_sector_map[ 16] = 0xffffff80;
-  bad_sector_map[ 17] = 0xffffffff;
-  bad_sector_map[ 18] = 0x0000ffff;
-#endif
-#if 0
-  /*  Enable to test bad sector handling
-   */
-  bad_sector_map[ 30] = 0xfffffffe;
-  bad_sector_map[ 32] = 0x7fffffff;
-  bad_sector_map[ 34] = 0xfffeffff;
-  bad_sector_map[ 36] = 0x55555555;
-  bad_sector_map[ 38] = 0xffffffff;
-  bad_sector_map[ 50] = 0xffff0000;
-  bad_sector_map[ 51] = 0xffffffff;
-  bad_sector_map[ 52] = 0xffffffff;
-  bad_sector_map[ 53] = 0x0000ffff;
-#endif
-#if 0
-  /*  Enable when testing multiple volume tar dumps.
-   */
-  for (i = first_data_segment; i <= ftape_last_segment.id - 7; ++i) {
-    bad_sector_map[ i] = EMPTY_SEGMENT;
-  }
-#endif
-#if 0
-  /*  Enable when testing bit positions in crc_error_map
-   */
-  for (i = first_data_segment; i <= ftape_last_segment.id; ++i) {
-    bad_sector_map[ i] |= 0x00ff00ff;
-  }
-#endif
-    
-  if (tracing > 2) {    
-    unsigned int map;
-    int good_sectors = 0;
-    int bad_sectors;
-    unsigned int total_bad = 0;
-
-    for (i = first_data_segment;
-         i < segments_per_track * tracks_per_tape; ++i) {
-      map = bad_sector_map[ i];
-      bad_sectors = count_ones( map);
-      if (bad_sectors > 0) {
-        TRACEx2( 6, "bsm for segment %4d: 0x%08x", i, map);
-        if (bad_sectors > SECTORS_PER_SEGMENT - 3) {
-          bad_sectors = SECTORS_PER_SEGMENT - 3;
-        }
-        total_bad += bad_sectors;
-      }
-      good_sectors += SECTORS_PER_SEGMENT - 3 - bad_sectors;
-    }
-    TRACEx1( 3, "%d Kb usable on this tape",
-            good_sectors - ftape_last_segment.free);
-    if (total_bad == 0) {
-      TRACE( 1, "WARNING: this tape has no bad blocks registered !");
-    }
-  }
   /* Copy the failed sector log into our local buffer.
    */
-  extract_file_marks( address);
+  if (!ftape_validate_label( &deblock_buffer[ 30])) {
+    TRACE( -1, "This tape has no `Linux raw format' label,\n"
+          "***** Use `mt' to erase this tape if you want to use file marks !");
+  } else {
+    extract_file_marks( address);
+  }
   ftape_reset_position();
   TRACE_EXIT;
   return 0;
@@ -539,20 +551,26 @@ int
 _ftape_read( char* buff, int req_len)
 {
   TRACE_FUN( 5, "_ftape_read");
-  int result;
+  int result = 0;
   int cnt;
   int to_do = req_len;
   static int remaining;
   int bytes_read = 0;
 
-  history.used |= 1;
-  if (first_data_segment == -1) {
-    result = read_header_segment( deblock_buffer);
-    if (result < 0) {
-      TRACE_EXIT;
-      return result;
+  if (ftape_offline || !formatted || no_tape) {
+    TRACEx3( -1, "offline = %d, formatted = %d, no_tape = %d",
+            ftape_offline, formatted, no_tape);
+    result = -EIO;
+  } else {
+    history.used |= 1;
+    if (first_data_segment == -1) {
+      result = read_header_segment( deblock_buffer);
     }
   }
+  if (result < 0) {
+    TRACE_EXIT;
+    return result;
+  }
   /*  As GNU tar doesn't accept partial read counts when the multiple
    *  volume flag is set, we make sure to return the requested amount
    *  of data. Except, of course, at the end of the tape or file mark.
@@ -602,7 +620,7 @@ _ftape_read( char* buff, int req_len)
     TRACEi( 7, "nr bytes just read:", cnt);
     result = verify_area( VERIFY_WRITE, buff, cnt);
     if (result) {
-      TRACE( 1, "verify_area failed");
+      TRACEx1( 1, "verify_area failed, exitcode = %d", result);
       TRACE_EXIT;
       return -EIO;
     }
