@@ -19,11 +19,11 @@ along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
  *
- $Source: /usr/src/distr/ftape-1.13a/RCS/ftape-read.h,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/ftape-read.h,v $
  $Author: bas $
  *
- $Revision: 1.7 $
- $Date: 1994/06/29 20:11:06 $
+ $Revision: 1.10 $
+ $Date: 1995/01/08 14:16:21 $
  $State: Beta $
  *
  *      This file contains the definitions for the read functions
@@ -43,5 +43,6 @@ extern int _ftape_read( char* buff, int req_len);
 extern int read_header_segment( byte* address);
 extern int read_segment( unsigned segment, byte* address, int* eof_mark,
                         int read_ahead);
+extern void ftape_zap_read_buffers( void);
 
 #endif /* _FTAPE_READ_H */
