@@ -15,11 +15,11 @@ You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
- $Source: /usr/src/distr/ftape-1.13a/RCS/ftape-write.c,v $
+ $Source: /home/bas/distr/ftape-2.02/RCS/ftape-write.c,v $
  $Author: bas $
  *
- $Revision: 1.9 $
- $Date: 1994/06/29 20:11:06 $
+ $Revision: 1.21 $
+ $Date: 1995/01/15 15:56:46 $
  $State: Beta $
  *
  *      This file contains the writing code
@@ -27,24 +27,23 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
 static char RCSid[] =
-"$Id: ftape-write.c,v 1.9 1994/06/29 20:11:06 bas Beta $";
+"$Id: ftape-write.c,v 1.21 1995/01/15 15:56:46 bas Beta bas $";
 
 
 #include <linux/string.h>
 #include <linux/errno.h>
-#include <linux/sched.h>
-#include <asm/dma.h>
 #include <asm/segment.h>
-#include <asm/system.h>
 
 #include "ftape.h"
 #include "ftape-write.h"
 #include "ftape-read.h"
 #include "qic117.h"
 #include "ftape-io.h"
+#include "ftape-ctl.h"
 #include "ftape-rw.h"
 #include "ftape-eof.h"
 #include "ecc.h"
+#include "ftape-bsm.h"
 
 
 /*      Global vars.
@@ -54,23 +53,20 @@ static char RCSid[] =
  */
 static int buf_pos_wr = 0;
 static int last_write_failed = 0;
-static bad_sector_map_changed = 0;
 static int need_flush = 0;
 
 #define WRITE_MULTI  0
 #define WRITE_SINGLE 1
 
 void
-add_segment_to_bad_sector_map( unsigned segment)
+ftape_zap_write_buffers( void)
 {
-  bad_sector_map[ segment] = EMPTY_SEGMENT;
-  bad_sector_map_changed = 1;
-}
+  int i;
 
-unsigned long
-get_bad_sector_map_entry( unsigned segment)
-{
-  return bad_sector_map[ segment];
+  for (i = 0; i < NR_ITEMS( buffer); ++i) {
+    buffer[ i].status = done;
+  }
+  need_flush = 0;
 }
 
 int
@@ -119,8 +115,8 @@ start_writing( int mode)
   buffer_struct* buff = &buffer[ head];
   int segment_id = buff->segment_id;
 
-  if (buff->status == full) {
-    setup_new_segment( buff, segment_id, 0, SECTORS_PER_SEGMENT);
+  if (ftape_state == writing && buff->status == waiting) {
+    setup_new_segment( buff, segment_id, 1);
     if (mode == WRITE_SINGLE) {
       buffer[ head].next_segment = 0; /* stop tape instead of pause */
     }
@@ -128,7 +124,7 @@ start_writing( int mode)
     buff->status = writing;
     if (runner_status == idle) {
       TRACEi( 5, "starting runner for segment", segment_id);
-      result= ftape_start_tape( segment_id); /* start moving */
+      result = ftape_start_tape( segment_id, buff->sector_offset);
       if (result >= 0) {
         runner_status = running;
       }
@@ -152,7 +148,7 @@ loop_until_writes_done( void)
   /*
    *  Wait until all data is actually written to tape.
    */
-  while (buffer[ head].status != empty) {
+  while (ftape_state == writing && buffer[ head].status != done) {
     TRACEx2( 7, "tail: %d, head: %d", tail, head);
     for (i = 0; i < NR_ITEMS( buffer); ++i) {
       TRACEx3( 7 /* 8 */, "buffer[ %d] segment_id: %d, status: %d",
@@ -173,14 +169,9 @@ loop_until_writes_done( void)
         result = -EINTR;          /* is this the right return value ? */
         break;
       }
-      if (buffer[ head].retry <= RETRIES_ON_CRC_ERROR) {
-        buffer[ head].status = full; /* retry this one */
-      } else {
-        result = -EIO;
-        break;
-      }
+      buffer[ head].status = waiting; /* retry this one */
       if (runner_status == aborting) {
-        ftape_smart_stop( &location_known, &current_segment);
+        ftape_dumb_stop();
         runner_status = idle;
       }
       if (runner_status != idle) {
@@ -210,15 +201,14 @@ write_segment( unsigned segment_id, byte* address, int flushing)
   if (ftape_state != writing) {
     if (ftape_state == reading) {
       TRACE( 5, "calling ftape_abort_operation");
-      result = ftape_abort_operation( &current_segment);
+      result = ftape_abort_operation();
       if (result < 0) {
         TRACE( 1, "ftape_abort_operation failed");
       }
     }
-    if (result >= 0) {
-      ftape_zap_buffers();        /* clear remaining read buffers */
-      ftape_state = writing;
-    }
+    ftape_zap_read_buffers();
+    ftape_zap_write_buffers();
+    ftape_state = writing;
   }
   /*    if all buffers full we'll have to wait...
    */
@@ -226,9 +216,16 @@ write_segment( unsigned segment_id, byte* address, int flushing)
   if (buffer[ tail].status == error) {
     /*  setup for a retry
      */
-    buffer[ tail].status = full;
+    buffer[ tail].status = waiting;
     bytes_written = -EAGAIN;    /* force retry */
-  } else if (buffer[ tail].status != empty) {
+    if (buffer[ tail].hard_error_map != 0) {
+      TRACEx1( 1, "warning: %d hard error(s) in written segment",
+              count_ones( buffer[ tail].hard_error_map));
+      TRACEx1( 4, "hard_error_map = 0x%08lx", buffer[ tail].hard_error_map);
+    }
+  } else if (buffer[ tail].status == done) {
+    history.defects += count_ones( buffer[ tail].hard_error_map);
+  } else {
     TRACE( 1, "wait for empty segment failed");
     result = -EIO;
   }
@@ -237,7 +234,7 @@ write_segment( unsigned segment_id, byte* address, int flushing)
   if (result >= 0 && runner_status == logical_eot) {
     int status;
 
-    result = ftape_ready_wait( 10 * SECOND, &status);
+    result = ftape_ready_wait( timeout.seek, &status);
     if (result < 0 || status & (QIC_STATUS_AT_BOT | QIC_STATUS_AT_EOT) == 0) {
       TRACE( 1, "eot/bot not reached");
     } else {
@@ -250,12 +247,12 @@ write_segment( unsigned segment_id, byte* address, int flushing)
       (runner_status == aborting || runner_status == buffer_underrun ||
        runner_status == end_of_tape)) {
     if (runner_status != end_of_tape) {
-      result = ftape_smart_stop( &location_known, &current_segment);
+      result = ftape_dumb_stop();
     }
     if (result >= 0) {
       if (runner_status == aborting) {
         if (buffer[ head].status == writing) {
-          buffer[ head].status = empty; /* ????? */
+          buffer[ head].status = done; /* ????? */
         }
       }
       runner_status = idle;       /* aborted ? */
@@ -264,19 +261,19 @@ write_segment( unsigned segment_id, byte* address, int flushing)
   /*  Don't start tape if runner idle and segment empty.
    */
   if (result >= 0 && !(runner_status == idle &&
-                       bad_sector_map[ segment_id] == EMPTY_SEGMENT)) {
-    if (buffer[ tail].status == empty) {
+                       get_bad_sector_entry( segment_id) == EMPTY_SEGMENT)) {
+    if (buffer[ tail].status == done) {
       /*    now at least one buffer is empty, fill it with our data.
        *    skip bad sectors and generate ecc.
        *    copy_and_gen_ecc return nr of bytes written,
        *    range 0..29 Kb inclusive !
        */
       result = copy_and_gen_ecc( buffer[ tail].address, address,
-                                bad_sector_map[ segment_id]);
+                                get_bad_sector_entry( segment_id));
       if (result >= 0) {
         bytes_written = result;
         buffer[ tail].segment_id = segment_id;
-        buffer[ tail].status = full;
+        buffer[ tail].status = waiting;
         next_buffer( &tail);
       }
     }
@@ -284,7 +281,7 @@ write_segment( unsigned segment_id, byte* address, int flushing)
      *    This will give higher probability of streaming.
      */
     if (result >= 0 && runner_status != running &&
-        ((head == tail && buffer[ tail].status == full) || flushing)) {
+        ((head == tail && buffer[ tail].status == waiting) || flushing)) {
       result = start_writing( WRITE_MULTI);
     }
   }
@@ -315,11 +312,12 @@ _write_segment( unsigned int segment_id, byte* buffer, int flush)
     result = write_segment( segment_id, buffer, flush);
     if (result < 0) {
       if (result == -EAGAIN) {
-        TRACEx1( 2, "write error, retrying (%d)", retry);
-        if (++retry > RETRIES_ON_WRITE_ERROR) {
-          TRACE( 1, "write error, all retries failed");
+        if (++retry > 100) {
+          TRACE( 1, "write failed, >100 retries in segment");
           result = -EIO;        /* give up */
           break;
+        } else {
+          TRACEx1( 2, "write error, retry %d", retry);
         }
       } else {
         TRACEi( 1, "write_segment failed, error:", -result);
@@ -354,10 +352,9 @@ update_header_segment( unsigned segment, byte* buffer)
   if (buffer == NULL) {
     TRACE( 5, "no input buffer specified");
     buffer = deblock_buffer;
-    result = read_segment( segment, buffer, &status, 0);
+    result = read_segment( used_header_segment, buffer, &status, 0);
     if (bad_sector_map_changed) {
-      memcpy( buffer + 2 * SECTOR_SIZE,
-             bad_sector_map, sizeof( bad_sector_map));
+      store_bad_sector_map( buffer);
     }
     if (failed_sector_log_changed) {
       update_failed_sector_log( buffer);
@@ -410,9 +407,6 @@ ftape_write_header_segments( byte* buffer)
   int header_1_ok = 0;
   int header_2_ok = 0;
 
-#if 0
-  ftape_state = idle;
-#endif
   do {
     if (!header_1_ok) {
       result = update_header_segment( header_segment_1, buffer);
@@ -421,7 +415,6 @@ ftape_write_header_segments( byte* buffer)
       }
       header_1_ok = 1;
     }
-#if 0
     if (!header_2_ok) {
       result = update_header_segment( header_segment_2, buffer);
       if (result < 0) {
@@ -429,8 +422,7 @@ ftape_write_header_segments( byte* buffer)
       }
       header_2_ok = 1;
     }
-#endif
-  } while (result < 0 && ++retry < 2 /* 6 */);
+  } while (result < 0 && retry++ < 3);
   if (result < 0) {
     if (!header_1_ok) {
       TRACE( 1, "update of first header segment failed");
@@ -456,26 +448,23 @@ ftape_update_header_segments( byte* buffer, int update)
     result = loop_until_writes_done();
   }
   if (result >= 0) {
-    result = ftape_abort_operation( &current_segment);
+    result = ftape_abort_operation();
     if (result >= 0) {
       if (buffer == NULL) {
         if (bad_sector_map_changed || failed_sector_log_changed) {
           ftape_seek_to_bot();  /* prevents extra rewind */
           buffer = deblock_buffer;
-          result = read_segment( header_segment_1, buffer, &dummy, 0);
+          result = read_segment( used_header_segment, buffer, &dummy, 0);
           if (result < 0) {
-            result = read_segment( header_segment_2, buffer, &dummy, 0);
-            if (result < 0) {
-              TRACE_EXIT;
-              return result;
-            }
+            TRACE_EXIT;
+            return result;
           }
         }
         header_changed = 0;
       }
       if (update) {
         if (bad_sector_map_changed) {
-          update_bad_sector_map( buffer);
+          store_bad_sector_map( buffer);
           header_changed = 1;
         }
         if (failed_sector_log_changed) {
@@ -486,18 +475,6 @@ ftape_update_header_segments( byte* buffer, int update)
       if (header_changed) {
         ftape_seek_to_bot();    /* prevents extra rewind */
         result = ftape_write_header_segments( buffer);
-        if (result >= 0) {
-          if (runner_status == idle) {
-            result = start_writing( WRITE_MULTI);
-          }
-          if (result >= 0) {
-            result = loop_until_writes_done();
-            if (update) {
-              bad_sector_map_changed =
-              failed_sector_log_changed = 0;
-            }
-          }
-        }
       }
     }
   }
@@ -534,7 +511,7 @@ ftape_flush_buffers( void)
   buf_pos_wr = 0;               /* prevent further writes if this fails */
   TRACE( 5, "flushing write buffers");
   if (last_write_failed) {
-    ftape_zap_buffers();
+    ftape_zap_write_buffers();
     active = 0;
     TRACE_EXIT;
     return write_protected ? -EROFS : -EIO;
@@ -560,7 +537,7 @@ ftape_flush_buffers( void)
         TRACE_EXIT;
         return result;
       } else {
-        clear_eof_mark( ftape_seg_pos, result);
+        clear_eof_mark_if_set( ftape_seg_pos, result);
       }
       TRACEi( 7, "flush, moved out buffer:", result);
       if (result < data_remaining) { /* need next segment for more data */
@@ -611,32 +588,22 @@ _ftape_write( char* buff, int req_len)
   int cnt;
   int written = 0;
 
+  if (write_protected) {
+    TRACE( 1, "error: cartridge write protected");
+    last_write_failed = 1;
+    result = -EROFS;
+  } else if (ftape_offline || !formatted || no_tape) {
+    result = -EIO;
+  } else if (first_data_segment == -1) {
   /*
    *    If we haven't read the header segment yet, do it now.
    *    This will verify the configuration, get the eof markers
    *    and the bad sector table.
    *    We'll use the deblock buffer for scratch.
    */
-  if (first_data_segment == -1) {
     result = read_header_segment( deblock_buffer);
-    if (result >= 0) {
-      if (strncmp( &deblock_buffer[ 30],
-                  LINUX_TAPE_LABEL, LINUX_TAPE_LABEL_LEN) != 0) {
-        TRACE( -1, "This tape has no `Linux raw format' label,\n"
-              "Erase it before further use !");
-        result = -EIO;
-      }
-    }
-  }
-  if (result >= 0) {
-    if (write_protected) {
-      last_write_failed = 1;
-      TRACE( 1, "error: cartridge write protected");
-      result = -EROFS;
-    } else {
-      if (result >= 0 && ftape_seg_pos > ftape_last_segment.id) {
-        result = -ENOSPC;       /* full is full */
-      }
+    if (result >= 0 && ftape_seg_pos > ftape_last_segment.id) {
+      result = -ENOSPC;         /* full is full */
     }
   }
   if (result < 0) {
@@ -693,7 +660,7 @@ _ftape_write( char* buff, int req_len)
         TRACE_EXIT;
         return result;
       } else {
-        clear_eof_mark( ftape_seg_pos, result);
+        clear_eof_mark_if_set( ftape_seg_pos, result);
       }
       if (result > 0 && result < buf_pos_wr) {
         /* Partial write: move remainder in lower part of buffer
@@ -742,16 +709,9 @@ ftape_fix( void)
     header_segment_2 = 1;
     first_data_segment = 2;
     result = read_segment( header_segment_2, scratch_buffer, &dummy, 0);
-    result = ftape_ready_wait( 5 * SECOND, &status);
+    result = ftape_ready_wait( timeout.pause, &status);
     result = ftape_write_header_segments( scratch_buffer);
   }
   TRACE_EXIT;
   return result;
 }
-
-
-
-
-
-
-
